/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  evm_ms_driver.c
 **
 **  Driver instance for the EVM M&S collector
 **/

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "gcspy_gc_stream.h"
#include "evm_ms_driver.h"

static ms_driver_tile_t *
msDriverAllocateStats (int tileNum) {
  int len = tileNum * sizeof(ms_driver_tile_t);
  ms_driver_tile_t *tiles = (ms_driver_tile_t *) malloc(len);
  if (tiles == NULL) {
    char buffer[256];
    sprintf(buffer, "M&S Driver: could not allocate %d bytes for tile data",
	    len);
    gcspy_raise_error(buffer);
  }
  return tiles;
}

static void
msDriverSetupTileNames (ms_driver_t *msDriver,
			int from,
			int to) {
  int i;
  char tmp[256];
  for (i = from; i < to; ++i) {
    gcspy_dUtilsRangeString(&(msDriver->area), i, tmp);
    gcspy_driverSetTileName(msDriver->driver, i, tmp);
  }
}

static ms_driver_tile_t *
msDriverGetTile (ms_driver_t *msDriver, int index) {
  return (ms_driver_tile_t *)
    gcspy_d_utils_get_stats_struct(&(msDriver->area), index, 0);
}

static int
msDriverGetTileIndex (ms_driver_t *msDriver, char *addr) {
  return gcspy_d_utils_get_index(&(msDriver->area), addr);
}

static char *
msDriverGetTileAddr (ms_driver_t *msDriver, int index) {
  return gcspy_d_utils_get_addr(&(msDriver->area), index);
}

void
msDriverZero (ms_driver_t *msDriver,
	      char *limit) {
  int i;
  ms_driver_tile_t *tile;
  int totalSpace = limit - msDriver->area.start;

  if (limit != msDriver->area.end) {
    int tileNum = gcspy_dUtilsTileNum(msDriver->area.start,
				      limit,
				      msDriver->area.blockSize); 
    /* printf("resizing to %8x, tileNum = %d\n", limit, tileNum); */
    msDriver->area.end = limit;
    msDriver->area.blockNum = tileNum;
    free(msDriver->area.stats);
    msDriver->area.stats = (char *) msDriverAllocateStats(tileNum);

    gcspy_driverResize(msDriver->driver, tileNum);
    msDriverSetupTileNames(msDriver, 0, tileNum);
  }

  for (i = 0; i < msDriver->area.blockNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    tile->usedSpace = 0;
    tile->cards = MS_CARD_STATE_CLEAN;
    tile->promotion = 0;
    tile->roots = 0;
    tile->marking = 0;
    tile->freeChunks = 0;
    tile->objects = 0;
  }

  msDriver->totalUsedSpace[0] = 0;
  msDriver->totalUsedSpace[1] = totalSpace;
  for (i = 0; i < 3; ++i) {
    msDriver->totalCardTable[i] = 0;
  }
  msDriver->totalPromotion[0] = 0;
  msDriver->totalPromotion[1] = totalSpace;
  msDriver->totalRoots = 0;
  msDriver->totalMarking = 0;
  msDriver->totalChunks = 0;
  msDriver->totalObjects = 0;
}

#if 0
void
msDriverLimit (ms_driver_t *msDriver,
	       char *limit) {
  int totalSpace = limit - msDriver->area.start;
  msDriver->totalUsedSpace[1] = totalSpace;
  msDriver->totalPromotion[1] = totalSpace;
  msDriver->limit = limit;
}
#endif //0

void
msDriverObject (ms_driver_t *msDriver,
		char *start,
		int size) {
  ++msDriver->totalObjects;
  msDriver->totalUsedSpace[0] += size;
  gcspy_dUtilsSetPerc(&(msDriver->area), start, start + size,
		      offsetof(ms_driver_tile_t, usedSpace));
  gcspy_dUtilsAddOne(&(msDriver->area), start, start + size,
		     offsetof(ms_driver_tile_t, objects));
}

void
msDriverFreeChunk (ms_driver_t *msDriver,
		   char *start,
		   int size) {
  ++msDriver->totalChunks;
  gcspy_dUtilsAddOne(&(msDriver->area), start, start + size,
		     offsetof(ms_driver_tile_t, freeChunks));
}

void
msDriverCard (ms_driver_t *msDriver,
	      char *start,
	      int size,
	      int state) {
  ++msDriver->totalCardTable[state];
  gcspy_dUtilsUpdateEnumDesc(&(msDriver->area), start, start + size,
			     offsetof(ms_driver_tile_t, cards), state);
}

void
msDriverPromoted (ms_driver_t *msDriver,
		  char *start,
		  int size) {
  msDriver->totalPromotion[0] += size;
  gcspy_dUtilsSetPerc(&(msDriver->area), start, start + size,
		      offsetof(ms_driver_tile_t, promotion));
}

void
msDriverRoot (ms_driver_t *msDriver,
	      char *start) {
  ++msDriver->totalRoots;
  gcspy_dUtilsAddSingle(&(msDriver->area), start, 
			offsetof(ms_driver_tile_t, roots));
}

void
msDriverMarked (ms_driver_t *msDriver,
		char *start) {
  ++msDriver->totalMarking;
  gcspy_dUtilsAddSingle(&(msDriver->area), start, 
			offsetof(ms_driver_tile_t, marking));
}

void
msDriverSend (ms_driver_t *msDriver, unsigned event) {
  int i;
  ms_driver_tile_t *tile;
  gcspy_gc_driver_t *driver = msDriver->driver;
  int tileNum = msDriver->area.blockNum;
  double perc;
  char tmp[128];
  int size;

  gcspy_driverStartComm(driver);

  gcspy_driverStream(driver, MS_USED_SPACE_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamIntValue(driver, tile->usedSpace);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_USED_SPACE_STREAM, 2);
  gcspy_driverSummaryValue(driver, msDriver->totalUsedSpace[0]);
  gcspy_driverSummaryValue(driver, msDriver->totalUsedSpace[1]);
  gcspy_driverSummaryEnd(driver);



  gcspy_driverStream(driver, MS_CARD_TABLE_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamByteValue(driver, tile->cards);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_CARD_TABLE_STREAM, 3);
  gcspy_driverSummaryValue(driver, msDriver->totalCardTable[0]);
  gcspy_driverSummaryValue(driver, msDriver->totalCardTable[1]);
  gcspy_driverSummaryValue(driver, msDriver->totalCardTable[2]);
  gcspy_driverSummaryEnd(driver);



  gcspy_driverStream(driver, MS_PROMOTION_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamIntValue(driver, tile->promotion);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_PROMOTION_STREAM, 2);
  gcspy_driverSummaryValue(driver, msDriver->totalPromotion[0]);
  gcspy_driverSummaryValue(driver, msDriver->totalPromotion[1]);
  gcspy_driverSummaryEnd(driver);



  gcspy_driverStream(driver, MS_ROOTS_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamShortValue(driver, tile->roots);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_ROOTS_STREAM, 1);
  gcspy_driverSummaryValue(driver, msDriver->totalRoots);
  gcspy_driverSummaryEnd(driver);



  gcspy_driverStream(driver, MS_MARKING_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamIntValue(driver, tile->marking);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_MARKING_STREAM, 1);
  gcspy_driverSummaryValue(driver, msDriver->totalMarking);
  gcspy_driverSummaryEnd(driver);



  gcspy_driverStream(driver, MS_CHUNKS_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamIntValue(driver, tile->freeChunks);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_CHUNKS_STREAM, 1);
  gcspy_driverSummaryValue(driver, msDriver->totalChunks);
  gcspy_driverSummaryEnd(driver);



  gcspy_driverStream(driver, MS_OBJECTS_STREAM, tileNum);
  for (i = 0; i < tileNum; ++i) {
    tile = msDriverGetTile(msDriver, i);
    gcspy_driverStreamIntValue(driver, tile->objects);
  }
  gcspy_driverStreamEnd(driver);

  gcspy_driverSummary(driver, MS_OBJECTS_STREAM, 1);
  gcspy_driverSummaryValue(driver, msDriver->totalObjects);
  gcspy_driverSummaryEnd(driver);



  size = msDriver->area.end - msDriver->area.start;
  sprintf(tmp, "Current Size: %s\n", gcspy_formatSize(size));
  gcspy_driverSpaceInfo(driver, tmp);

  gcspy_driverEndComm(driver);
}

void
msDriverInit (ms_driver_t *msDriver,
	      gcspy_gc_driver_t *gcDriver,
	      const char *name,
	      unsigned blockSize,
	      char *start,
	      char *end) {
  int i;
  char tmp[256];
  gcspy_gc_stream_t *stream;
  int tileNum = gcspy_dUtilsTileNum(start, end, blockSize);
  ms_driver_tile_t *tiles = msDriverAllocateStats(tileNum);

  msDriver->driver = gcDriver;
  gcspy_dUtilsInit(&(msDriver->area),
		   start, end,
		   0, blockSize, tileNum,
		   (char *) tiles, sizeof(ms_driver_tile_t));

  if (blockSize < 1024)
    sprintf(tmp, "Block Size: %d bytes\n", blockSize);
  else
    sprintf(tmp, "Block Size: %dK\n", (blockSize / 1024));
  gcspy_driverInit(gcDriver, -1, name, "M&S GC",
		   "Block ", tmp, tileNum, NULL, 1);
  msDriverSetupTileNames(msDriver, 0, tileNum);

  stream = gcspy_driverAddStream(gcDriver, MS_USED_SPACE_STREAM);
  gcspy_streamInit(stream, MS_USED_SPACE_STREAM,
		   GCSPY_GC_STREAM_INT_TYPE,
		   "Used Space",
		   0, blockSize,
		   0, 0,
		   "Used Space: ", " bytes",
		   GCSPY_GC_STREAM_PRESENTATION_PERCENT,
		   GCSPY_GC_STREAM_PAINT_STYLE_ZERO, 0,
		   gcspy_colorDBGetColorWithName("Red"));

  stream = gcspy_driverAddStream(gcDriver, MS_CARD_TABLE_STREAM);
  gcspy_streamInit(stream, MS_CARD_TABLE_STREAM,
		   GCSPY_GC_STREAM_BYTE_TYPE,
		   "Card Table",
		   0, 2,
		   2, 2,
		   "Card State: ", "",
		   GCSPY_GC_STREAM_PRESENTATION_ENUM,
		   GCSPY_GC_STREAM_PAINT_STYLE_PLAIN, 0,
		   gcspy_colorDBGetColorWithName("Off White"));
  gcspy_streamAddEnumName(stream, MS_CARD_STATE_DIRTY, "DIRTY");
  gcspy_streamAddEnumName(stream, MS_CARD_STATE_SUMMARISED, "SUMMARISED");
  gcspy_streamAddEnumName(stream, MS_CARD_STATE_CLEAN, "CLEAN");

  stream = gcspy_driverAddStream(gcDriver, MS_PROMOTION_STREAM);
  gcspy_streamInit(stream, MS_PROMOTION_STREAM,
		   GCSPY_GC_STREAM_INT_TYPE,
		   "Promotion",
		   0, blockSize,
		   0, 0,
		   "Promoted: ", " bytes",
		   GCSPY_GC_STREAM_PRESENTATION_PERCENT,
		   GCSPY_GC_STREAM_PAINT_STYLE_ZERO, 0,
		   gcspy_colorDBGetColorWithName("Magenta"));

  stream = gcspy_driverAddStream(gcDriver, MS_ROOTS_STREAM);
  gcspy_streamInit(stream, MS_ROOTS_STREAM,
		   GCSPY_GC_STREAM_SHORT_TYPE,
		   "Roots",
		   0, gcspy_d_utils_roots_per_block(blockSize),
		   0, 0,
		   "Roots: ", "",
		   GCSPY_GC_STREAM_PRESENTATION_MAX_VAR,
		   GCSPY_GC_STREAM_PAINT_STYLE_ZERO, 0,
		   gcspy_colorDBGetColorWithName("Green"));

  stream = gcspy_driverAddStream(gcDriver, MS_MARKING_STREAM);
  gcspy_streamInit(stream, MS_MARKING_STREAM,
		   GCSPY_GC_STREAM_INT_TYPE,
		   "Marking",
		   0, gcspy_d_utils_objects_per_block(blockSize),
		   0, 0,
		   "Marked: ", " objects",
		   GCSPY_GC_STREAM_PRESENTATION_PLUS,
		   GCSPY_GC_STREAM_PAINT_STYLE_ZERO, 0,
		   gcspy_colorDBGetColorWithName("Orange"));

  stream = gcspy_driverAddStream(gcDriver, MS_CHUNKS_STREAM);
  gcspy_streamInit(stream, MS_CHUNKS_STREAM,
		   GCSPY_GC_STREAM_INT_TYPE,
		   "Free Chunks",
		   0, gcspy_d_utils_chunks_per_block(blockSize),
		   0, 0,
		   "Free Chunks: ", "",
		   GCSPY_GC_STREAM_PRESENTATION_PLUS,
		   GCSPY_GC_STREAM_PAINT_STYLE_ZERO, 0,
		   gcspy_colorDBGetColorWithName("Blue"));

  stream = gcspy_driverAddStream(gcDriver, MS_OBJECTS_STREAM);
  gcspy_streamInit(stream, MS_OBJECTS_STREAM,
		   GCSPY_GC_STREAM_INT_TYPE,
		   "Objects",
		   0, gcspy_d_utils_objects_per_block(blockSize),
		   0, 0,
		   "Objects: ", "",
		   GCSPY_GC_STREAM_PRESENTATION_PLUS,
		   GCSPY_GC_STREAM_PAINT_STYLE_ZERO, 0,
		   gcspy_colorDBGetColorWithName("Yellow"));
}
