/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  gcspy_main_server.cpp
 **
 **  Main server that runs inside the visualised system
 **/

#include "gcspy_main_server.hpp"

gcspyMainServer::gcspyMainServer (int port,
				  const char *name,
				  gcspy_length_t max_len,
				  gcspy_bool_t verbose) {
  _max_len         = max_len;
  _space_num       = 0;
  _event_num       = 0;
  _connected       = 0;
  _shutdown_now    = 0;
  _shutdown_sent   = 0;
  _verbose         = verbose;
  _server_state    = new gcspyServerState();
  if (_verbose) {
    char buffer[128];
    sprintf(buffer, "starting server, port %d", port);
    verbose_print(buffer);
  }
  if (_max_len > 0)
    _server = new gcspyServer(port, max_len);
  else
    _server = new gcspyServer(port);
  _client = new gcspyClient();
  _interpreter = new gcspyInterpreter(_client);
  gcspyUtils::copy_string(_name, name, name_len);
  set_general_info("");
}

void
gcspyMainServer::add_space (gcspySpace *space) {
  gcspy_assert(_space_num < max_spaces);
  space->set_id(_space_num);
  _spaces[_space_num] = space;
  ++_space_num;
}

void
gcspyMainServer::add_event (gcspy_index_t index, char *event_name) {
  gcspy_assert(_event_num < gcspyServerState::max_events);
  gcspy_assert(_event_num == index);
  char *str = gcspyUtils::clone_string(event_name);
  if (str == NULL) {
    char buffer[128];
    sprintf(buffer, "new failed for event %s (%d)\n", event_name, index);
    gcspy_raise_error(buffer);
  }
  _event_names[index] = event_name;
  _event_counts[index] = 0;
  ++_event_num;
}

/**
 **    Initial checks when serialised
 **    ------------------------------
 **      magic_num          string
 **      endianess          int
 **      pause_request      boolean (only client sends this)
 **      server_name        string (only server sends this)
 **/

gcspy_bool_t
gcspyMainServer::check_client () {
  char buffer[128];

  _client->definitely_receive();
  gcspyBufferedInput input = gcspyBufferedInput(_client);
  input.read_string(buffer);
  if (strcmp(buffer, magic_string))
    return 0;
  if (input.read_int() != endianess)
    return 0;
  if (input.read_boolean())
    _pause_now = 1;
  if (!input.has_finished())
    return 0;
  input.close();

  gcspyBufferedOutput output = gcspyBufferedOutput(_client);
  output.write_string(magic_string);
  output.write_int(endianess);
  output.write_string(_name);
  output.close();
  output.send();

  return 1;
}

/**
 **    Boot info format when serialised
 **    ----------------------------------
 **      space_num        short
 **      general_info     string
 **      event_num        short
 **      ( event_name )*  string
 **    ----------------------------------
 **    then call each space to serialise itself
 **    ----------------------------------
 **/

void
gcspyMainServer::send_boot_info () {
  int i;

  gcspyBufferedOutput output = gcspyBufferedOutput(_client);
  output.write_short((short) _space_num);
  output.write_string(_general_info);
  output.write_short((short) _event_num);
  for (i = 0; i < _event_num; ++i)
    output.write_string(_event_names[i]);
  output.close();
  output.send();

  for (i = 0; i < _space_num; ++i)
    _spaces[i]->serialise(_client);
}

void
gcspyMainServer::internal_setup () {
  _shutdown_now = 0;
  _shutdown_sent = 0;
  _paused = 0;
  _pause_now = 0;
}

gcspy_bool_t
gcspyMainServer::connect () {
  verbose_print("waiting for client to connect");
  _client->wait_for_connection(_server, _max_len);
  internal_setup();
  verbose_print("  client connected");

  if (!check_client())
    return 0;

  send_boot_info();
  _server_state->reset_event_filters();

  reset_compensation_timer();
  start_elapsed_timer();
  _connected = 1;

  return 1;
}

gcspy_bool_t
gcspyMainServer::internal_loop () {
  gcspy_bool_t shutdown_req = 0;

  verbose_print("  starting server main loop");
  while ( 1 ) {
    _client->receive();
    if (_client->has_finished()) {
      verbose_print("  main loop terminated (client died)");
      _connected = 0;
      break;
    }

    _server_state->reset_flags();
    gcspyBufferedInput input = gcspyBufferedInput(_client);
    _interpreter->execute(&input, (void *) _server_state);
    if (_server_state->is_shutdown_req_set()) {
      verbose_print("  main loop terminated (shutdown request received)");
      shutdown_req = 1;
      break;
    }
    if (_server_state->is_pause_req_set()) {
      _pause_now = 1;
    }
    if (_server_state->is_restart_set()) {
      _paused = 0;
    }
  }

  return shutdown_req;
}

void
gcspyMainServer::wait_for_client () {
  verbose_print("blocked until client connects");

  while (!is_connected()) {
    gcspyTimer::wait(50);
  }
  safepoint();
}

void
gcspyMainServer::transmit_streams () {
  for (int i = 0; i < _space_num; ++i) {
    gcspySpace *space = _spaces[i];
    gcspy_length_t tile_num = space->get_tile_num();
    if (space->has_changed()) {
      _interpreter->send_space_cmd(space);
      space->unflag_changed();
    }

    for (int j = 0; j < space->get_stream_num(); ++j) {
      gcspyStream *stream = space->get_stream(j);
      _interpreter->send_stream_cmd(i, j, stream->get_data_type(),
				    stream->get_data(), tile_num);
      int *summary = stream->get_summary();
      if (summary != NULL) {
	_interpreter->send_summary_cmd(i, j, summary,
				       stream->get_summary_len());
      }
    }
    _interpreter->send_space_info_cmd(i, space->get_space_info());
    _interpreter->send_control_cmd(i, space->get_control(), tile_num);
  }
}

void
gcspyMainServer::main_loop () {
  gcspy_bool_t shutdown_req;
  while ( 1 ) {
    connect();

    shutdown_req = internal_loop();
    if (shutdown_req) {
      _shutdown_now = 1;
      while (!_shutdown_sent) {
	gcspyTimer::wait(50);
      }
      _connected = 0;
    }

    _client->close();
  }
}

void
gcspyMainServer::internal_safepoint () {
  if (_shutdown_now) {
    _interpreter->send_shutdown_cmd();
    _shutdown_sent = 1;
  }

  if (_pause_now) {
    verbose_print("  paused");
    _interpreter->send_pause_cmd();
    _pause_now = 0;
    _paused = 1;
  }

  if (_paused) {
    while ( _paused && is_connected() ) {
      gcspyTimer::wait(10);
      if (_shutdown_now) {
	_interpreter->send_shutdown_cmd();
	_shutdown_sent = 1;
	_paused = 0;
      }
      if (_server_state->is_play_one_set()) {
	verbose_print("  play one");
	_server_state->unset_play_one();
	break;
      }
    }

    if (!_paused && !_shutdown_now) {
      verbose_print("  restarted");
    }
  }
}
