/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  hs_mc_driver.cpp
 **
 **  Abstract driver for the HS old and permanent generation collectors
 **/

#include "hs_mc_driver.hpp"

char *
hsMCDriver::generate_block_info (gcspy_size_t block_size) {
  static char buffer[128];
  if (block_size < 1024)
    sprintf(buffer, "Block Size: %d bytes\n", block_size);
  else
    sprintf(buffer, "Block Size: %dK\n", (block_size / 1024));
    return buffer;
}

void
hsMCDriver::retrieve_data_arrays () {
  _used_space_data      = _used_space_stream->get_int_data();
  _dead_space_data      = _dead_space_stream->get_int_data();
  _card_table_data      = _card_table_stream->get_byte_data();
  _object_data          = _object_stream->get_int_data();
  _marking_data         = _marking_stream->get_int_data();
  _refs_data            = _refs_stream->get_int_data();
  _refs_to_young_data   = _refs_to_young_stream->get_int_data();
  _internal_refs_data   = _internal_refs_stream->get_int_data();
}

void
hsMCDriver::init_summaries (gcspy_size_t area_size) {
  _used_space_summary[1]   = area_size;
  _dead_space_summary[1]   = area_size;
}

void
hsMCDriver::reset (char *end) {
  if (end != _area->get_end()) {
    _area->resize(end);
    resize(_area->get_block_num());
    retrieve_data_arrays();
    _area->setup_range_strings(this);
  }

  reset_data();
  gcspy_size_t area_size = _area->get_area_size();
  init_summaries(area_size);

  char buffer[256], tmp[64];
  sprintf(buffer, "Current Size: %s\nCard Size: %d bytes\n",
	  gcspyUtils::format_size(area_size, tmp),
	  _card_size);
  set_space_info(buffer);
}

hsMCDriver::hsMCDriver (char *name,
			gcspy_size_t block_size,
			gcspy_size_t card_size,
			char *start,
			char *end)
  : gcspySpace(name, "M&C Driver", "Block ",
	       generate_block_info(block_size),
	       gcspyContArea::calc_block_num(start, end, block_size),
	       NULL, 1) {
  _card_size = card_size;
  _area = new gcspyContArea(start, end, 0, block_size);
  _area->setup_range_strings(this);
}

void
hsMCDriver::add_used_space_stream () {
  _used_space_stream = new gcspyStream(gcspyStream::int_type,
				       "Used Space",
				       0, _area->get_block_size(), 0, 0,
				       "Used Space: ", " bytes",
				       gcspyStream::presentation_percent,
				       gcspyStream::paint_style_zero,
				       0, gcspyColorDB::get_color("Red"), 1);
  add_stream(_used_space_stream);
  _used_space_summary = _used_space_stream->get_summary();
}

void
hsMCDriver::add_dead_space_stream () {
  _dead_space_stream = new gcspyStream(gcspyStream::int_type,
				       "Dead Space",
				       0, _area->get_block_size(), 0, 0,
				       "Dead Space: ", " bytes",
				       gcspyStream::presentation_percent,
				       gcspyStream::paint_style_zero,
				       0, gcspyColorDB::get_color("Red"), 1);
  add_stream(_dead_space_stream);
  _dead_space_summary = _dead_space_stream->get_summary();
}

void
hsMCDriver::add_card_table_stream () {
  _card_table_stream = new gcspyStream(gcspyStream::byte_type,
				   "Card Table",
				   card_state_dirty,
				   card_state_clean,
				   0,
				   card_state_clean,
				   "Card State: ", "",
				   gcspyStream::presentation_enum,
				   gcspyStream::paint_style_plain,
				   0, gcspyColorDB::get_color("Off White"), 1);
  _card_table_stream->add_enum_name(card_state_dirty,   "DIRTY");
  _card_table_stream->add_enum_name(card_state_younger, "YOUNGER");
  _card_table_stream->add_enum_name(card_state_clean,   "CLEAN");
  add_stream(_card_table_stream);
  _card_table_summary = _card_table_stream->get_summary();
}

void
hsMCDriver::add_object_stream () {
  _object_stream = new gcspyStream(gcspyStream::int_type,
				  "Objects",
				  0,
		     gcspyContArea::objects_per_block(_area->get_block_size()),
				  0, 0,
				  "Objects: ", "",
				  gcspyStream::presentation_plus,
				  gcspyStream::paint_style_zero,
				  0, gcspyColorDB::get_color("Yellow"), 1);
  add_stream(_object_stream);
  _object_summary = _object_stream->get_summary();
}

void
hsMCDriver::add_marking_stream () {
  _marking_stream = new gcspyStream(gcspyStream::int_type,
			          "Marking",
			          0,
		     gcspyContArea::objects_per_block(_area->get_block_size()),
				  0, 0,
				  "Marked: ", " objects",
				  gcspyStream::presentation_plus,
				  gcspyStream::paint_style_zero,
				  0, gcspyColorDB::get_color("Orange"), 1);
  add_stream(_marking_stream);
  _marking_summary = _marking_stream->get_summary();
}

void
hsMCDriver::add_refs_stream () {
  _refs_stream = new gcspyStream(gcspyStream::int_type,
				 "Reference Fields",
				 0,
			gcspyContArea::refs_per_block(_area->get_block_size()),
				 0, 0,
				 "Reference Fields: ", "",
				 gcspyStream::presentation_plus,
				 gcspyStream::paint_style_zero,
				 0, gcspyColorDB::get_color("Green"), 1);
  add_stream(_refs_stream);
  _refs_summary = _refs_stream->get_summary();
}

void
hsMCDriver::add_refs_to_young_stream () {
  _refs_to_young_stream = new gcspyStream(gcspyStream::int_type,
				     "Reference Fields (Young)",
				     0,
			gcspyContArea::refs_per_block(_area->get_block_size()),
				     0, 0,
				     "Reference Fields: ", " (Young)",
				     gcspyStream::presentation_plus,
				     gcspyStream::paint_style_zero,
				     0, gcspyColorDB::get_color("Green"), 1);
  add_stream(_refs_to_young_stream);
  _refs_to_young_summary = _refs_to_young_stream->get_summary();
}

void
hsMCDriver::add_internal_refs_stream () {
  _internal_refs_stream = new gcspyStream(gcspyStream::int_type,
				     "Internal References",
				     0,
			gcspyContArea::refs_per_block(_area->get_block_size()),
				     0, 0,
				     "Internal References: ", "",
				     gcspyStream::presentation_plus,
				     gcspyStream::paint_style_zero,
				     0, gcspyColorDB::get_color("Green"), 1);
  add_stream(_internal_refs_stream);
  _internal_refs_summary = _internal_refs_stream->get_summary();

  retrieve_data_arrays();
}
